/*
* SACD Decoder plugin
* Copyright (c) 2011-2015 Maxim V.Anisiutkin <maxim.anisiutkin@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with FFmpeg; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

#pragma once

#include <windows.h>
#include <iasiodrv.h>
#include <foobar2000.h>
#include "audio_device.h"
#include "asio_util.h"

#define MAXDRVNAMELEN 128

enum asio_state_e { ASIO_STATE_UNLOADED, ASIO_STATE_LOADED, ASIO_STATE_INITIALIZED, ASIO_STATE_PREPARED, ASIO_STATE_RUNNING };

class asio_device_t : public audio_device_t {
	CRITICAL_SECTION m_cs;
	GUID             m_guid;
	bool             m_trace;
	IASIO*                        m_pCoASIO;
	asio_state_e                  m_state;
	t_size                        m_channels;
	t_size                        m_samples;
	bool                          m_can_DSD;
	bool                          m_run_DSD;
	char                          m_name[MAXDRVNAMELEN];
	long                          m_version;
	long                          m_numInputChannels;
	long                          m_numOutputChannels;
	long                          m_minSize;
	long                          m_maxSize;
	long                          m_preferredSize;
	long                          m_granularity;
	long                          m_inputLatency;
	long                          m_outputLatency;
	pfc::array_t<ASIOChannelInfo> m_dsd_channelInfos;
	pfc::array_t<ASIOChannelInfo> m_pcm_channelInfos;
	pfc::array_t<ASIOBufferInfo>  m_bufferInfos;
public:
	static void enumerate(output_device_enum_callback& p_callback);
	asio_device_t();
	~asio_device_t();
	GUID get_GUID() {
		return m_guid;
	}
	IASIO* get_pIASIO() {
		return m_pCoASIO;
	}
	asio_state_e get_state() {
		return m_state;
	}
	t_size get_channels() {
		return m_channels;
	}
	t_size get_samples() {
		return m_samples;
	}
	bool get_can_DSD() {
		return m_can_DSD;
	}
	bool get_run_DSD() {
		return m_run_DSD;
	}
	const char* get_name() {
		return m_name;
	}
	pfc::array_t<ASIOChannelInfo>& get_channelInfos(bool p_DSD) {
		return p_DSD ? m_dsd_channelInfos : m_pcm_channelInfos;
	}
	pfc::array_t<ASIOBufferInfo>& get_bufferInfos() {
		return m_bufferInfos;
	}
	bool load(const GUID& p_GUID);
	bool unload();
	bool configure();
	bool can_time_info();
	bool can_report_overload();
	bool can_do_format(ASIOIoFormatType p_format);
	bool set_run_DSD(bool p_DSD);
	ASIOIoFormatType get_format();
	bool get_latencies();
	bool get_sample_position(t_uint64& p_position, t_uint64& p_timestamp);
	unsigned get_samplerate();
	bool set_samplerate(unsigned p_samplerate);
	bool create_buffers(ASIOCallbacks* p_callbacks);
	bool dispose_buffers();
	bool start();
	bool stop();
	bool output_ready();
private:
	bool init(void* sysHandle);
	bool set_format(ASIOIoFormatType p_fomat);
	bool set_channels();
	bool set_channelInfos();
	bool set_samples();
	bool future(long selector, void* opt);
};

