/*
* SACD Decoder plugin
* Copyright (c) 2011-2015 Maxim V.Anisiutkin <maxim.anisiutkin@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with FFmpeg; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

#include "asio_util.h"

long asio_util_t::getSampleSizeInBits(ASIOSampleType sampleType) {
	long size;
	switch (sampleType) {
	case ASIOSTDSDInt8MSB1:
	case ASIOSTDSDInt8LSB1:
		size = 1;
		break;
	case ASIOSTDSDInt8NER8:
		size = 8;
		break;
	case ASIOSTInt16MSB:
	case ASIOSTInt16LSB:
		size = 16;
		break;
	case ASIOSTInt24MSB:
	case ASIOSTInt24LSB:
		size = 24;
		break;
	case ASIOSTInt32MSB:
	case ASIOSTInt32MSB16:
	case ASIOSTInt32MSB18:
	case ASIOSTInt32MSB20:
	case ASIOSTInt32MSB24:
	case ASIOSTInt32LSB:
	case ASIOSTInt32LSB16:
	case ASIOSTInt32LSB18:
	case ASIOSTInt32LSB20:
	case ASIOSTInt32LSB24:
	case ASIOSTFloat32MSB:
	case ASIOSTFloat32LSB:
		size = 32;
		break;
	case ASIOSTFloat64MSB:
	case ASIOSTFloat64LSB:
		size = 64;
		break;
	default:
		size = 0;
		break;
	}
	return size;
}

long asio_util_t::getSamplePrecision(ASIOSampleType sampleType) {
	long size;
	switch (sampleType) {
	case ASIOSTDSDInt8MSB1:
	case ASIOSTDSDInt8LSB1:
		size = 1;
		break;
	case ASIOSTDSDInt8NER8:
		size = 8;
		break;
	case ASIOSTInt16MSB:
	case ASIOSTInt16LSB:
	case ASIOSTInt32MSB16:
	case ASIOSTInt32LSB16:
		size = 16;
		break;
	case ASIOSTInt32MSB18:
	case ASIOSTInt32LSB18:
		size = 18;
		break;
	case ASIOSTInt32MSB20:
	case ASIOSTInt32LSB20:
		size = 18;
		break;
	case ASIOSTInt24MSB:
	case ASIOSTInt24LSB:
	case ASIOSTInt32MSB24:
	case ASIOSTInt32LSB24:
		size = 24;
		break;
	case ASIOSTInt32MSB:
	case ASIOSTInt32LSB:
	case ASIOSTFloat32MSB:
	case ASIOSTFloat32LSB:
		size = 32;
		break;
	case ASIOSTFloat64MSB:
	case ASIOSTFloat64LSB:
		size = 64;
		break;
	default:
		size = 0;
		break;
	}
	return size;
}

const char* asio_util_t::getSampleTypeName(ASIOSampleType sampleType) {
	switch (sampleType) {
	case ASIOSTDSDInt8MSB1:
		return "DSDInt8MSB1";
	case ASIOSTDSDInt8LSB1:
		return "DSDInt8LSB1";
	case ASIOSTDSDInt8NER8:
		return "DSDInt8NER8";
	case ASIOSTInt16MSB:
		return "Int16MSB";
	case ASIOSTInt16LSB:
		return "Int16LSB";
	case ASIOSTInt24MSB:
		return "Int24MSB";
	case ASIOSTInt24LSB:
		return "Int24LSB";
	case ASIOSTInt32MSB:
		return "Int32MSB";
	case ASIOSTInt32MSB16:
		return "Int32MSB16";
	case ASIOSTInt32MSB18:
		return "Int32MSB18";
	case ASIOSTInt32MSB20:
		return "Int32MSB20";
	case ASIOSTInt32MSB24:
		return "Int32MSB24";
	case ASIOSTInt32LSB:
		return "Int32LSB";
	case ASIOSTInt32LSB16:
		return "Int32LSB16";
	case ASIOSTInt32LSB18:
		return "Int32LSB18";
	case ASIOSTInt32LSB20:
		return "Int32LSB20";
	case ASIOSTInt32LSB24:
		return "Int32LSB24";
	case ASIOSTFloat32MSB:
		return "Float32MSB";
	case ASIOSTFloat32LSB:
		return "Float32LSB";
	case ASIOSTFloat64MSB:
		return "Float64MSB";
	case ASIOSTFloat64LSB:
		return "Float64LSB";
	}
	return "Unknown";
}

bool asio_util_t::isMSB(ASIOSampleType sampleType) {
	bool msb;
	switch (sampleType) {
	case ASIOSTDSDInt8MSB1:
	case ASIOSTDSDInt8NER8:
	case ASIOSTInt16MSB:
	case ASIOSTInt24MSB:
	case ASIOSTInt32MSB:
	case ASIOSTInt32MSB16:
	case ASIOSTInt32MSB18:
	case ASIOSTInt32MSB20:
	case ASIOSTInt32MSB24:
	case ASIOSTFloat32MSB:
	case ASIOSTFloat64MSB:
		msb = true;
		break;
	default:
		msb = false;
		break;
	}
	return msb;
}

bool asio_util_t::isFP(ASIOSampleType sampleType) {
	bool fp;
	switch (sampleType) {
	case ASIOSTFloat32MSB:
	case ASIOSTFloat32LSB:
	case ASIOSTFloat64MSB:
	case ASIOSTFloat64LSB:
		fp = true;
		break;
	default:
		fp = false;
		break;
	}
	return fp;
}

const char* asio_util_t::getFutureSelectorName(long selector) {
	switch (selector) {
	case kAsioEnableTimeCodeRead:
		return "kAsioEnableTimeCodeRead";
	case kAsioDisableTimeCodeRead:
		return "kAsioDisableTimeCodeRead";
	case kAsioSetInputMonitor:
		return "kAsioSetInputMonitor";
	case kAsioTransport:
		return "kAsioTransport";
	case kAsioSetInputGain:
		return "kAsioSetInputGain";
	case kAsioGetInputMeter:
		return "kAsioGetInputMeter";
	case kAsioSetOutputGain:
		return "kAsioSetOutputGain";
	case kAsioGetOutputMeter:
		return "kAsioGetOutputMeter";
	case kAsioCanInputMonitor:
		return "kAsioCanInputMonitor";
	case kAsioCanTimeInfo:
		return "kAsioCanTimeInfo";
	case kAsioCanTimeCode:
		return "kAsioCanTimeCode";
	case kAsioCanTransport:
		return "kAsioCanTransport";
	case kAsioCanInputGain:
		return "kAsioCanInputGain";
	case kAsioCanInputMeter:
		return "kAsioCanInputMeter";
	case kAsioCanOutputGain:
		return "kAsioCanOutputGain";
	case kAsioCanOutputMeter:
		return "kAsioCanOutputMeter";
	case kAsioOptionalOne:
		return "kAsioOptionalOne";
	case kAsioSetIoFormat:
		return "kAsioSetIoFormat";
	case kAsioGetIoFormat:
		return "kAsioGetIoFormat";
	case kAsioCanDoIoFormat:
		return "kAsioCanDoIoFormat";
	case kAsioCanReportOverload:
		return "kAsioCanReportOverload";
	case kAsioGetInternalBufferSamples:
		return "kAsioGetInternalBufferSamples";
	default:
		return "<Unknown Selector>";
	}
}

const char* asio_util_t::getMessageSelectorName(long selector) {
	switch (selector) {
	case kAsioSelectorSupported:
		return "kAsioSelectorSupported";
	case kAsioEngineVersion:
		return "kAsioEngineVersion";
	case kAsioResetRequest:
		return "kAsioResetRequest";
	case kAsioBufferSizeChange:
		return "kAsioBufferSizeChange";
	case kAsioResyncRequest:
		return "kAsioResyncRequest";
	case kAsioLatenciesChanged:
		return "kAsioLatenciesChanged";
	case kAsioSupportsTimeInfo:
		return "kAsioSupportsTimeInfo";
	case kAsioSupportsTimeCode:
		return "kAsioSupportsTimeCode";
	case kAsioMMCCommand:
		return "kAsioMMCCommand";
	case kAsioSupportsInputMonitor:
		return "kAsioSupportsInputMonitor";
	case kAsioSupportsInputGain:
		return "kAsioSupportsInputGain";
	case kAsioSupportsInputMeter:
		return "kAsioSupportsInputMeter";
	case kAsioSupportsOutputGain:
		return "kAsioSupportsOutputGain";
	case kAsioSupportsOutputMeter:
		return "kAsioSupportsOutputMeter";
	case kAsioOverload:
		return "kAsioOverload";
	default:
		return "<Unknown Selector>";
	}
}

const char* asio_util_t::getIoFormatName(ASIOIoFormatType	format) {
	switch (format) {
	case kASIOFormatInvalid:
		return "kASIOFormatInvalid";
	case kASIOPCMFormat:
		return "kASIOPCMFormat";
	case kASIODSDFormat:
		return "kASIODSDFormat";
	default:
		return "<Unknown Format>";
	}
}

const char* asio_util_t::getASIOError(ASIOError error) {
	switch (error) {
	case ASE_OK:
		return "ASE_OK";
	case ASE_SUCCESS:
		return "ASE_SUCCESS";
	case ASE_NotPresent:
		return "ASE_NotPresent";
	case ASE_HWMalfunction:
		return "ASE_HWMalfunction";
	case ASE_InvalidParameter:
		return "ASE_InvalidParameter";
	case ASE_InvalidMode:
		return "ASE_InvalidMode";
	case ASE_SPNotAdvancing:
		return "ASE_SPNotAdvancing";
	case ASE_NoClock:
		return "ASE_NoClock";
	case ASE_NoMemory:
		return "ASE_NoMemory";
	default:
		return "<Unknown Error>";
	}
};
