/*
* SACD Decoder plugin
* Copyright (c) 2011-2015 Maxim V.Anisiutkin <maxim.anisiutkin@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with FFmpeg; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

#include "audio_config.h"
#include "asio_device.h"
#include "wasapi_device.h"

audio_config_t g_audio_config;

static const GUID output_id_dsd = { 0xf4d74794, 0x2936, 0x4595,{ 0x8d, 0x6a, 0xc0, 0xce, 0x7c, 0x5b, 0x35, 0x25 } };

static const GUID g_guid_cfg_device_guid = { 0x68177cc8, 0x64a2, 0x42f9,{ 0xb6, 0x80, 0x17, 0x5f, 0x6b, 0x66, 0x85, 0x97 } };
static cfg_guid g_cfg_device_guid(g_guid_cfg_device_guid, GUID_NULL);

static const GUID g_guid_cfg_DoP = { 0xa510da7c, 0xafbf, 0x4e70,{ 0xb8, 0x31, 0x14, 0x2c, 0xec, 0x77, 0x58, 0xb7 } };
static cfg_objList<device_DoP_t> g_cfg_DoP(g_guid_cfg_DoP);

static const GUID g_guid_cfg_channel_mapping = { 0x9e739d2c, 0xf334, 0x4775,{ 0x83, 0x1b, 0x45, 0xc9, 0x75, 0xb8, 0x1d, 0x90 } };
static cfg_objList<device_channel_mapping_t> g_cfg_channel_mapping(g_guid_cfg_channel_mapping);

static const GUID g_guid_cfg_trace = { 0x21260974, 0x5312, 0x4c3f,{ 0x85, 0x18, 0x32, 0x64, 0xca, 0x6f, 0x73, 0x37 } };
static cfg_int g_cfg_trace(g_guid_cfg_trace, 0);

GUID audio_config_t::g_get_guid() {
	return output_id_dsd;
}

const char* audio_config_t::g_get_name() {
	return "DSD";
}

bool audio_config_t::g_get_trace() {
	return g_cfg_trace != 0;
}

bool audio_config_t::g_set_trace(bool p_trace) {
	bool modified = false;
	if ((g_cfg_trace != 0) != p_trace) {
		g_cfg_trace = p_trace;
		modified = true;
	}
	return modified;
}

audio_config_t::audio_config_t() {
	m_channels = 0;
	m_DSD = false;
	m_DoP = false;
	enum_devices();
}

audio_config_t::~audio_config_t() {
}

t_size audio_config_t::get_device_count() {
	return m_devices.get_count();
}

unsigned audio_config_t::get_current_device() {
	for (unsigned item = 0; item < m_devices.get_count(); item++) {
		if (m_devices[item].m_guid == g_cfg_device_guid) {
			select_device(item);
			return item;
		}
	}
	select_device(0);
	return 0;
}

void audio_config_t::set_current_device(unsigned p_item) {
	if (p_item < get_device_count()) {
		g_cfg_device_guid = m_devices[p_item].m_guid;
		select_device(p_item);
	}
}

bool audio_config_t::can_do_DSD() {
	return m_DSD;
}

bool audio_config_t::get_DoP() {
	return m_DoP;
}

bool audio_config_t::set_DoP(bool p_DoP) {
	bool modified = false;
	if (m_DoP != p_DoP) {
		m_DoP = p_DoP;
		modified = true;
	}
	return modified;
}

const char* audio_config_t::get_device_name(unsigned p_item) {
	return (p_item < get_device_count()) ? m_devices[p_item].m_name.get_ptr() : nullptr;
}

t_size audio_config_t::get_channel_count() {
	return m_channels;
}

unsigned audio_config_t::get_channel_mapping(unsigned p_channel) {
	unsigned inp_channel;
	inp_channel = 1 << p_channel;
	if (p_channel < m_channel_mapping.get_count()) {
		inp_channel = m_channel_mapping[p_channel].m_inp_channel;
	}
	else {
		inp_channel = 1 << p_channel;
	}
	return inp_channel;
}

bool audio_config_t::set_channel_mapping(unsigned p_channel, unsigned p_inp_channel) {
	bool modified = false;
	if (p_channel < m_channel_mapping.get_count()) {
		if (m_channel_mapping[p_channel].m_inp_channel != p_inp_channel) {
			m_channel_mapping[p_channel].m_inp_channel = p_inp_channel;
			modified = true;
		}
	}
	return modified;
}

const char* audio_config_t::get_inp_channel_name(unsigned p_inp_channel) {
	switch (p_inp_channel) {
	case audio_chunk::channel_front_left:
		return "Left";
	case audio_chunk::channel_front_right:
		return "Right";
	case audio_chunk::channel_front_center:
		return "Center";
	case audio_chunk::channel_lfe:
		return "LFE";
	case audio_chunk::channel_back_left:
		return "Surround Left";
	case audio_chunk::channel_back_right:
		return "Surround Right";
	case audio_chunk::channel_front_center_left:
		return "Front Center Left";
	case audio_chunk::channel_front_center_right:
		return "Front Center Right";
	case audio_chunk::channel_back_center:
		return "Back Center";
	case audio_chunk::channel_side_left:
		return "Side Left";
	case audio_chunk::channel_side_right:
		return "Side Right";
	case audio_chunk::channel_top_center:
		return  "Top Center";
	case audio_chunk::channel_top_front_left:
		return "Top Front Left";
	case audio_chunk::channel_top_front_center:
		return "Top Front Center";
	case audio_chunk::channel_top_front_right:
		return "Top Front Right";
	case audio_chunk::channel_top_back_left:
		return "Top Surround Left";
	case audio_chunk::channel_top_back_center:
		return "Top Surround Center";
	case audio_chunk::channel_top_back_right:
		return "Top Surround Right";
	default:
		return "<none>";
	}
}

const char* audio_config_t::get_out_channel_name(bool p_DSD, unsigned p_channel) {
	const char* name = nullptr;
	if (p_DSD) {
		if (p_channel < m_out_dsd_channel_names.get_count()) {
			name = m_out_dsd_channel_names[p_channel].get_ptr();
		}
	}
	else {
		if (p_channel < m_out_pcm_channel_names.get_count()) {
			name = m_out_pcm_channel_names[p_channel].get_ptr();
		}
	}
	return name;
}

const char* audio_config_t::get_out_channel_sample_type(bool p_DSD, unsigned p_channel) {
	const char* type = nullptr;
	if (p_DSD) {
		if (p_channel < m_out_dsd_channel_sample_types.get_count()) {
			type = m_out_dsd_channel_sample_types[p_channel].get_ptr();
		}
	}
	else {
		if (p_channel < m_out_pcm_channel_sample_types.get_count()) {
			type = m_out_pcm_channel_sample_types[p_channel].get_ptr();
		}
	}
	return type;
}

void audio_config_t::load_DoP() {
	m_DoP = false;
	for (unsigned cfg_item = 0; cfg_item < g_cfg_DoP.get_count(); cfg_item++) {
		if (g_cfg_DoP[cfg_item].m_key == g_cfg_device_guid) {
			m_DoP = g_cfg_DoP[cfg_item].m_DoP;
		}
	}
}

void audio_config_t::save_DoP() {
	bool add_item = true;
	for (unsigned cfg_item = 0; cfg_item < g_cfg_DoP.get_count(); cfg_item++) {
		if (g_cfg_DoP[cfg_item].m_key == g_cfg_device_guid) {
			g_cfg_DoP[cfg_item].m_DoP = m_DoP;
			add_item = false;
		}
	}
	if (add_item) {
		device_DoP_t DoP;
		DoP.m_key = g_cfg_device_guid;
		DoP.m_DoP = m_DoP;
		g_cfg_DoP.add_item(DoP);
	}
}

void audio_config_t::load_channel_mapping(t_size p_channels) {
	m_channel_mapping.set_count(p_channels);
	for (unsigned out_channel = 0; out_channel < m_channel_mapping.get_count(); out_channel++) {
		m_channel_mapping[out_channel].m_inp_channel = 1 << out_channel;
		m_channel_mapping[out_channel].m_out_channel = out_channel;
	}
	for (unsigned cfg_item = 0; cfg_item < g_cfg_channel_mapping.get_count(); cfg_item++) {
		if (g_cfg_channel_mapping[cfg_item].m_key == g_cfg_device_guid) {
			if (g_cfg_channel_mapping[cfg_item].m_channel_mapping.m_out_channel < m_channel_mapping.get_count()) {
				m_channel_mapping[g_cfg_channel_mapping[cfg_item].m_channel_mapping.m_out_channel] = g_cfg_channel_mapping[cfg_item].m_channel_mapping;
			}
		}
	}
}

void audio_config_t::save_channel_mapping() {
	for (unsigned out_channel = 0; out_channel < m_channel_mapping.get_count(); out_channel++) {
		bool add_item = true;
		for (unsigned cfg_item = 0; cfg_item < g_cfg_channel_mapping.get_count(); cfg_item++) {
			if (g_cfg_channel_mapping[cfg_item].m_key == g_cfg_device_guid) {
				if (g_cfg_channel_mapping[cfg_item].m_channel_mapping.m_out_channel == out_channel) {
					g_cfg_channel_mapping[cfg_item].m_channel_mapping.m_inp_channel = m_channel_mapping[out_channel].m_inp_channel;
					add_item = false;
				}
			}
		}
		if (add_item) {
			device_channel_mapping_t ch_map;
			ch_map.m_key = g_cfg_device_guid;
			ch_map.m_channel_mapping = m_channel_mapping[out_channel];
			g_cfg_channel_mapping.add_item(ch_map);
		}
	}
}

bool audio_config_t::select_device(unsigned p_item) {
	bool ok = false;
	m_channels = 0;
	m_out_dsd_channel_names.set_count(m_channels);
	m_out_dsd_channel_sample_types.set_count(m_channels);
	m_out_pcm_channel_names.set_count(m_channels);
	m_out_pcm_channel_sample_types.set_count(m_channels);
	GUID guid = GUID_NULL;
	if (p_item < get_device_count()) {
		guid = m_devices[p_item].m_guid;
		if (g_asio_device_playback.get_GUID() == guid && g_asio_device_playback.get_state() >= ASIO_STATE_INITIALIZED) {
			m_DSD = g_asio_device_playback.get_can_DSD();
			m_channels = g_asio_device_playback.get_channelInfos(m_DSD && !m_DoP).get_count();
			load_DoP();
			load_channel_mapping(m_channels);
			if (m_DSD) {
				m_out_dsd_channel_names.set_count(m_channels);
				m_out_dsd_channel_sample_types.set_count(m_channels);
				for (unsigned out_channel = 0; out_channel < m_channels; out_channel++) {
					ASIOChannelInfo info = g_asio_device_playback.get_channelInfos(true)[out_channel];
					m_out_dsd_channel_names[out_channel] = info.name;
					m_out_dsd_channel_sample_types[out_channel] = asio_util_t::getSampleTypeName(info.type);
				}
			}
			m_out_pcm_channel_names.set_count(m_channels);
			m_out_pcm_channel_sample_types.set_count(m_channels);
			for (unsigned out_channel = 0; out_channel < m_channels; out_channel++) {
				ASIOChannelInfo info = g_asio_device_playback.get_channelInfos(false)[out_channel];
				m_out_pcm_channel_names[out_channel] = info.name;
				m_out_pcm_channel_sample_types[out_channel] = asio_util_t::getSampleTypeName(info.type);
			}
			ok = true;
		}
		else {
			if (m_asio_device.load(guid)) {
				m_DSD = m_asio_device.get_can_DSD();
				m_channels = m_asio_device.get_channelInfos(m_DSD && !m_DoP).get_count();
				load_DoP();
				load_channel_mapping(m_channels);
				if (m_DSD) {
					m_out_dsd_channel_names.set_count(m_channels);
					m_out_dsd_channel_sample_types.set_count(m_channels);
					for (unsigned out_channel = 0; out_channel < m_channels; out_channel++) {
						ASIOChannelInfo info = m_asio_device.get_channelInfos(true)[out_channel];
						m_out_dsd_channel_names[out_channel] = info.name;
						m_out_dsd_channel_sample_types[out_channel] = asio_util_t::getSampleTypeName(info.type);
					}
				}
				m_out_pcm_channel_names.set_count(m_channels);
				m_out_pcm_channel_sample_types.set_count(m_channels);
				for (unsigned out_channel = 0; out_channel < m_channels; out_channel++) {
					ASIOChannelInfo info = m_asio_device.get_channelInfos(false)[out_channel];
					m_out_pcm_channel_names[out_channel] = info.name;
					m_out_pcm_channel_sample_types[out_channel] = asio_util_t::getSampleTypeName(info.type);
				}
				ok = true;
			}
			m_asio_device.unload();
		}
	}
	return ok;
}

void audio_config_t::configure_device() {
	GUID guid = g_cfg_device_guid;
	if (g_asio_device_playback.get_GUID() == guid && g_asio_device_playback.get_state() >= ASIO_STATE_INITIALIZED) {
		g_asio_device_playback.configure();
	}
	else {
		if (m_asio_device.load(guid)) {
			m_asio_device.configure();
		}
		m_asio_device.unload();
	}
}

void audio_config_t::enum_devices() {
	enum_devices(*this);
}

void audio_config_t::enum_devices(output_device_enum_callback& p_callback) {
	m_devices.set_count(0);
	asio_device_t::enumerate(p_callback);
	wasapi_device_t::enumerate(p_callback);
}

void audio_config_t::reset() {
	g_cfg_DoP.set_count(0);
	g_cfg_channel_mapping.set_count(0);
	enum_devices();
	set_current_device(0);
	g_set_trace(false);
}

void audio_config_t::on_device(const GUID& p_guid, const char* p_name, unsigned p_name_length) {
	m_devices.increase_size(1);
	device_properties_t& props = m_devices[m_devices.get_count() - 1];
	props.m_guid = p_guid;
	props.m_name = p_name;
}
