/*
* SACD Decoder plugin
* Copyright (c) 2011-2015 Maxim V.Anisiutkin <maxim.anisiutkin@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with FFmpeg; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

#pragma once

#include <windows.h>
#include <foobar2000.h>
#include "audio_devices.h"

class device_property_t {
public:
	GUID m_key;
};

class device_DoP_t : public device_property_t {
public:
	bool m_DoP;
};

FB2K_STREAM_READER_OVERLOAD(device_DoP_t) {
	device_DoP_t temp;
	stream >> temp.m_key;
	stream >> temp.m_DoP;
	value = temp;
	return stream;
}

FB2K_STREAM_WRITER_OVERLOAD(device_DoP_t) {
	stream << value.m_key;
	stream << value.m_DoP;
	return stream;
}

class channel_mapping_t {
public:
	unsigned m_inp_channel;
	unsigned m_out_channel;
};

class device_channel_mapping_t : public device_property_t {
public:
	channel_mapping_t m_channel_mapping;
};

FB2K_STREAM_READER_OVERLOAD(device_channel_mapping_t) {
	device_channel_mapping_t temp;
	stream >> temp.m_key;
	stream >> temp.m_channel_mapping.m_inp_channel;
	stream >> temp.m_channel_mapping.m_out_channel;
	value = temp;
	return stream;
}

FB2K_STREAM_WRITER_OVERLOAD(device_channel_mapping_t) {
	stream << value.m_key;
	stream << value.m_channel_mapping.m_inp_channel;
	stream << value.m_channel_mapping.m_out_channel;
	return stream;
}

class device_properties_t {
public:
	GUID         m_guid;
	pfc::string8 m_name;
};

class audio_config_t : public output_device_enum_callback {
	asio_device_t&                    m_asio_device = g_asio_device_config;
	pfc::array_t<device_properties_t> m_devices;
	unsigned                          m_channels;
	pfc::array_t<pfc::string8>        m_out_dsd_channel_names;
	pfc::array_t<pfc::string8>        m_out_dsd_channel_sample_types;
	pfc::array_t<pfc::string8>        m_out_pcm_channel_names;
	pfc::array_t<pfc::string8>        m_out_pcm_channel_sample_types;
	bool                              m_DSD;
	bool                              m_DoP;
	pfc::array_t<channel_mapping_t>   m_channel_mapping;
public:
	static GUID g_get_guid();
	static const char* g_get_name();
	static bool g_get_trace();
	static bool g_set_trace(bool p_trace);
	audio_config_t();
	virtual ~audio_config_t();
	t_size get_device_count();
	unsigned get_current_device();
	void set_current_device(unsigned p_item);
	bool can_do_DSD();
	bool get_DoP();
	bool set_DoP(bool p_DoP);
	t_size get_channel_count();
	const char* get_device_name(unsigned p_item);
	unsigned get_channel_mapping(unsigned p_channel);
	bool set_channel_mapping(unsigned p_channel, unsigned p_inp_channel);
	const char* get_inp_channel_name(unsigned p_inp_channel);
	const char* get_out_channel_name(bool p_DSD, unsigned p_channel);
	const char* get_out_channel_sample_type(bool p_DSD, unsigned p_channel);
	bool select_device(unsigned p_item);
	void load_DoP();
	void save_DoP();
	void load_channel_mapping(t_size p_channels);
	void save_channel_mapping();
	void configure_device();
	void enum_devices();
	void enum_devices(output_device_enum_callback& p_callback);
	void reset();
	virtual void on_device(const GUID& p_guid, const char* p_name, unsigned p_name_length);
};

extern audio_config_t g_audio_config;